with Ada.Unchecked_Deallocation;

package body Paskma.Lists is

   procedure Free is 
   new Ada.Unchecked_Deallocation(Cell, Cell_Access);

   ------------------------------------------------------------------------------

   procedure Nullify (
         The_List : in out List ) is 
   begin
      The_List.Count := 0;
      The_List.Head := null;
      The_List.Tail := null;
   end Nullify;

   ------------------------------------------------------------------------------

   procedure Initialize (
         The_List : in out List ) is 
   begin
      Nullify(The_List);
   end;

   ------------------------------------------------------------------------------

   procedure Finalize (
         The_List : in out List ) is 
   begin
      Empty(The_List);
   end;

   ------------------------------------------------------------------------------

   procedure Adjust (
         The_List : in out List ) is 

      Temp_List :          List;  
      I         :          Iterator := Get_Begin (The_List);  
      I_End     : constant Iterator := Get_End (The_List);  
   begin
      --
      -- create the copy of "the_list"
      while I /= I_End loop
         Push_Back(Dereference(I), Temp_List);
         Increment(I);
      end loop;

      --
      -- "shallow" assigment
      The_List.Head := Temp_List.Head;
      The_List.Tail := Temp_List.Head;

      if The_List.Count /= Temp_List.Count then
         -- assertion
         raise Program_Error;
      else
         The_List.Count := Temp_List.Count;
      end if;

      Nullify(Temp_List);
      -- make temp_list don't release its cells
   end;

   ------------------------------------------------------------------------------

   function Get_Begin (
         Of_List : in     List ) 
     return Iterator is 
   begin

      if Of_List.Count = 0 then

         return Iterator'(
            Actual   => null, 
            Previous => null, 
            Is_End   => True);
      end if;

      return Iterator'(
         Actual   => Of_List.Head, 
         Previous => null,         
         Is_End   => False);
   end;

   ------------------------------------------------------------------------------

   function Length (
         Of_The_List : in     List ) 
     return Natural is 
   begin
      return Of_The_List.Count;
   end;

   ------------------------------------------------------------------------------

   function Get_End (
         Of_List : in     List ) 
     return Iterator is 
   begin

      return Iterator'(
         Actual   => null,         
         Previous => Of_List.Tail, 
         --null when list is empty
         Is_End => True);
   end;

   ------------------------------------------------------------------------------

   procedure Push_Front (
         Item        : in     Item_Type; 
         To_The_List : in out List       ) is 

      New_Cell : Cell_Access;  
   begin
      if To_The_List.Count = 0 then

         New_Cell := new Cell'(
            Value    => Item, 
            Previous => null, 
            Next     => null);

         To_The_List.Head := New_Cell;
         To_The_List.Tail := New_Cell;
         To_The_List.Count := 1;

      else
         New_Cell := new Cell'(
            Value    => Item,            
            Previous => null,            
            Next     => To_The_List.Head);

         To_The_List.Head.Previous := New_Cell;
         To_The_List.Head := New_Cell;

         To_The_List.Count := To_The_List.Count + 1;
      end if;

   end;

   ------------------------------------------------------------------------------

   procedure Push_Back (
         Item        : in     Item_Type; 
         To_The_List : in out List       ) is 

      New_Cell : Cell_Access;  
   begin
      if To_The_List.Count = 0 then

         New_Cell := new Cell'(
            Value    => Item, 
            Previous => null, 
            Next     => null);

         To_The_List.Head := New_Cell;
         To_The_List.Tail := New_Cell;
         To_The_List.Count := 1;

      else
         New_Cell := new Cell'(
            Value    => Item,             
            Previous => To_The_List.Tail, 
            Next     => null);

         To_The_List.Tail.Next := New_Cell;
         To_The_List.Tail := New_Cell;

         To_The_List.Count := To_The_List.Count + 1;
      end if;
   end;

   ------------------------------------------------------------------------------

   procedure Empty (
         The_List : in out List ) is 

      Cell_To_Delete,  
      Temp_Cell      : Cell_Access;  
   begin

      Temp_Cell := The_List.Head;

      while Temp_Cell /= null loop
         Cell_To_Delete := Temp_Cell;
         Temp_Cell := Temp_Cell.Next;
         Free(Cell_To_Delete);
      end loop;

      Nullify(The_List);

   end Empty;

   ------------------------------------------------------------------------------

   procedure Remove (
         I         : in out Iterator; 
         From_List : in out List      ) is 

      To_Delete : Cell_Access;  
   begin
      if  I.Is_End = True then -- if i.actual = get_end(from_list)
         raise Iterator_Error;
      end if;

      if From_List.Count = 0 then
         --defensive check, maybe, the Program_Error would be better
         raise List_Empty;
      end if;

      To_Delete := I.Actual;

      if From_List.Count = 1 then
         -- removing the last cell

         I := Iterator'( -- lets iterator to be invalid but "end"
            Actual   => null, 
            Previous => null, 
            Is_End   => True);

      elsif I.Actual = From_List.Head then
         -- removing the first cell
         From_List.Head := I.Actual.Next; -- we have a new head
         From_List.Head.Previous := null; -- head cell has not previous cell

         I := Get_Begin(From_List); -- lets iterator to be valid

      elsif I.Actual = From_List.Tail then
         -- removing the last cell
         From_List.Tail := I.Previous; -- we have a new tail
         From_List.Tail.Next := null; -- tail cell hes not next cell

         I := Get_End(From_List);  -- lets iterator to be valid
      else
         -- ordinary case
         I.Actual.Previous.Next := I.Actual.Next;
         I.Actual.Next.Previous := I.Actual.Previous;

         Increment(I);
      end if;

      Free(To_Delete);
      From_List.Count := From_List.Count - 1;

      if From_List.Count = 0 then
         -- set to the initial state
         Nullify(From_List);
      end if;
   end;


   ------------------------------------------------------------------------------

   function "=" (
         Left,                   
         Right : in     Iterator ) 
     return Boolean is 
   begin
      if Left.Is_End = True and Right.Is_End = True then
         return True;
      end if;

      return Left.Actual = Right.Actual and Left.Previous = Right.Previous;
   end;

   ------------------------------------------------------------------------------

   procedure Increment (
         I : in out Iterator ) is 
   begin
      if I.Actual = null then
         raise Iterator_Error;
      end if;

      I.Previous := I.Actual;
      I.Actual := I.Actual.Next;

      if I.Actual = null then
         I.Is_End := True;
      end if;
   end;

   ------------------------------------------------------------------------------

   procedure Decrement (
         I : in out Iterator ) is 
   begin
      if I.Previous = null then
         raise Iterator_Error;
      end if;

      I.Actual := I.Previous;
      I.Previous := I.Previous.Previous;

      I.Is_End := False;
   end Decrement;

   ------------------------------------------------------------------------------

   function Dereference (
         I : in     Iterator ) 
     return Item_Type is 
   begin
      if I.Actual = null then
         raise Iterator_Error;
      end if;

      return I.Actual.Value;
   end;

   ------------------------------------------------------------------------------

   procedure Dereference (
         I     : in out Iterator; 
         Value : in     Item_Type ) is 
   begin
      if I.Actual = null then
         raise Iterator_Error;
      end if;

      I.Actual.Value := Value;
   end;

   ------------------------------------------------------------------------------

   function Access_Dereference (
         I : in     Iterator ) 
     return Item_Access is 
   begin
      if I.Actual = null then
         raise Iterator_Error;
      end if;

      return I.Actual.Value'access;
   end Access_Dereference;

   ------------------------------------------------------------------------------

end Paskma.Lists;

