with Gnat.Os_Lib,
   Ada.Strings;

use Gnat.Os_Lib,
   Ada.Strings;

package body File_Utilities is

   ------------------
   -- Compose_Path --
   ------------------

   function Compose_Path (
         Directory,                   
         File      : Unbounded_String ) 
     return Unbounded_String is 
      -- file should be also directory path 

      Internal_File : Unbounded_String;  
   begin
      if Length(Directory) = 0 then
         return File;
      end if;

      --
      -- remove the initial directory_separator from the File string
      if Length(File) > 0 and then Element(File, 1) = Directory_Separator then
         Internal_File := Tail(File, Length(File) -1);
      else
         Internal_File := File;
      end if;

      --
      -- Note: if the file is "" then normalized directory is returned.
      if Element(Directory, Length(Directory)) = Directory_Separator then
         return Directory & Internal_File;
      else
         return Directory & Directory_Separator & Internal_File;
      end if;
   end Compose_Path;

   --
   -- Compose_Logical_Path
   --
   -- The directory separator is hard set to '/', this is logical file-tree _
   --  representation for FTP. Even if it is running on Windows

   function Compose_Logical_Path (
         Directory,                   
         File      : Unbounded_String ) 
     return Unbounded_String is 

      Internal_File : Unbounded_String;  
   begin
      if Length(Directory) = 0 then
         return File;
      end if;
      --
      -- remove the initial directory_separator from the File string
      if Length(File) > 0 and then Element(File, 1) = '/' then
         Internal_File := Tail(File, Length(File) -1);
      else
         Internal_File := File;
      end if;

      --
      -- Note: if the file is "" then normalized directory is returned.
      if Element(Directory, Length(Directory)) = '/' then
         return Directory & Internal_File;
      else
         return Directory & '/' & Internal_File;
      end if;
   end Compose_Logical_Path;

   --
   -- Normalize_Directory_Name
   --
   -- connecting the directory separator at and of the directory name if needed

   function Normalize_Directory_Name (
         Directory_Name : Unbounded_String ) 
     return Unbounded_String is 
   begin
      if Element(Directory_Name, Length(Directory_Name)) /= Directory_Separator then
         -- if there is not directory separator at the end then add one
         return Directory_Name & Directory_Separator;
      else
         -- else return unchanged
         return Directory_Name;
      end if;
   end Normalize_Directory_Name;

   --
   -- Display_Directory_Name 
   --
   -- cutts of the ending slash or back-slash (if any)
   -- if it represents the unix root directory, "/" is leaved

   function Display_Directory_Name (
         Directory_Name : Unbounded_String ) 
     return Unbounded_String is 
   begin
      if Directory_Name = To_Unbounded_String("/") then
         -- leave unix root directory sign unchanged
         return Directory_Name;
      elsif Element(Directory_Name, Length(Directory_Name)) = Directory_Separator
            or Element(Directory_Name, Length(Directory_Name)) = '/' then
         return Head(Directory_Name, Length(Directory_Name) -1 );
      else
         return Directory_Name;
      end if;
   end;

   --
   -- Get_First_Path_Element
   --
   -- For Logical Paths only
   -- Example: Get_First_Path_Element("/eins/zwei/drei") = "eins"

   function Get_First_Path_Element (
         Path : Unbounded_String ) 
     return Unbounded_String is 

      Internal_Path   : Unbounded_String;  
      Separator_Index : Natural;  
   begin
      if Length(Path) = 0 then
         raise Constraint_Error;
      end if;
      --
      -- cut initial slash (if any)
      if Element(Path, 1) = '/' then
         Internal_Path := Tail(Path, Length(Path)-1);
      else
         Internal_Path := Path;
      end if;

      Separator_Index := Index(Internal_Path, "/");

      if Separator_Index = 0 then
         -- there is not separator, so there is at most one path element
         -- return the one element
         return Internal_Path;
      end if;

      return Head(Internal_Path, Separator_Index-1);
   end;

   --
   -- Cut_First_Path_Element
   --
   -- For Logical Paths only
   -- Example: Cut_First_Path_Element("/eins/zwei/drei") = "zwei/drei"

   function Cut_First_Path_Element (
         Path : Unbounded_String ) 
     return Unbounded_String is 

      Internal_Path   : Unbounded_String;  
      Separator_Index : Natural;  
   begin
      if Length(Path) = 0 then
         raise Constraint_Error;
      end if;
      --
      -- cut the initial slahs (if any)
      if Element(Path, 1) = '/' then
         Internal_Path := Tail(Path, Length(Path)-1);
      else
         Internal_Path := Path;
      end if;

      Separator_Index := Index(Internal_Path, "/");

      if Separator_Index = 0 then
         -- there is not separator, so there is at most one path element
         -- return the ""
         return Null_Unbounded_String;
      end if;

      return Tail(Internal_Path, Length(Internal_Path)-Separator_Index);
      -- the initial slash is cutted (because we don't want to create absolute path)
   end;


   --
   -- Logical_To_Physical_Separator
   --

   function Logical_To_Physical_Separator (
         Filename : Unbounded_String ) 
     return Unbounded_String is 

      Result : Unbounded_String := Filename;  
   begin
      for I in 1..Length(Result) loop
         if Element(Result, I) = '/' then
            Replace_Element(Result, I, Directory_Separator);
         end if;
      end loop;

      return Result;
   end Logical_To_Physical_Separator;

   --
   -- cut_last_path_element
   --
   -- logical paths only

   function Cut_Last_Path_Element (
         Path : Unbounded_String ) 
     return Unbounded_String is 

      Internal_Path   : Unbounded_String;  
      Separator_Index : Natural;  
   begin
      if Length(Path) = 0 then
         raise Constraint_Error;
      end if;

      if Path = "/" then
         -- there is no element to cut
         return Null_Unbounded_String;
      end if;

      --
      -- cut the final slash (if any)
      if Element(Path, Length(Path)) = '/' then
         Internal_Path := Head(Path, Length(Path)-1);
      else
         Internal_Path := Path;
      end if;

      Separator_Index := Index(Internal_Path, "/", Backward);

      -- there should be always initial slash, so separator_index is never 0, but if
      -- the given path is relative then it is possible 

      if Separator_Index = 0 then
         -- there is not separator, so there is at most one path element
         -- return the ""
         return Null_Unbounded_String;
      end if;

      return Head(Internal_Path, Separator_Index);
      -- the final slash is leaved
   end;

   --
   -- cut_last_physical_path_element
   --
   -- physical paths only

   function Cut_Last_Physical_Path_Element (
         Path : Unbounded_String ) 
     return Unbounded_String is 

      Internal_Path   : Unbounded_String;  
      Separator_Index : Natural;  
   begin
      if Length(Path) = 0 then
         raise Constraint_Error;
      end if;

      if Path = "" & Directory_Separator then
         -- there is no element to cut
         return Null_Unbounded_String;
      end if;

      --
      -- cut the final (back)slash (if any)
      if Element(Path, Length(Path)) = Directory_Separator then
         Internal_Path := Head(Path, Length(Path)-1);
      else
         Internal_Path := Path;
      end if;

      Separator_Index := Index(Internal_Path, "" & Directory_Separator, Backward);

      if Separator_Index = 0 then
         -- there is not separator, so there is at most one path element
         -- return the ""
         return Null_Unbounded_String;
      end if;

      return Head(Internal_Path, Separator_Index);
      -- the final (back)slash is leaved
   end;

   --
   -- Get...
   --
   function Get_Last_Physical_Path_Element (
         Path : Unbounded_String ) 
     return Unbounded_String is 

      Internal_Path   : Unbounded_String;  
      Separator_Index : Natural;  
   begin
      if Length(Path) = 0 then
         raise Constraint_Error;
      end if;

      if Path = "" & Directory_Separator then
         -- there is no element to cut
         return Null_Unbounded_String;
      end if;

      --
      -- cut the final (back)slash (if any)
      if Element(Path, Length(Path)) = Directory_Separator then
         Internal_Path := Head(Path, Length(Path)-1);
      else
         Internal_Path := Path;
      end if;

      Separator_Index := Index(Internal_Path, "" & Directory_Separator, Backward);

      if Separator_Index = 0 then
         return Internal_Path;
      end if;

      return Tail(Internal_Path, Length(Internal_Path) - Separator_Index);
      -- the final (back)slash is leaved
   end;



end File_Utilities;

