with Ada.Text_Io,
   Ada.Exceptions,
   Paskma.String_Utilities,
   Access_Rights,
   File_Utilities,
   Paskma.Logger;

use Ada.Text_Io,
   Ada.Exceptions,
   Paskma.String_Utilities,
   Access_Rights,
   File_Utilities,
   Paskma.Logger;

package body Ftp_Configurations is

   use   Shared_Dir_Lists,  
   Mapping_Lists,  
   User_Conf_Lists;  

   --
   -- get_account
   --
   -- private method for Verify_user and Get_File
   procedure Get_Account (
         From_Conf    : in     Ftp_Configuration_Type;  
         Result       : in out User_Configuration_Type; 
         Of_Username,                                   
         And_Password :        Unbounded_String         ) is 

      I     :          User_Conf_Lists.Iterator := Get_Begin (From_Conf.Accounts);  
      I_End : constant User_Conf_Lists.Iterator := Get_End (From_Conf.Accounts);  
   begin
      Test_Loaded(From_Conf);

      while I /= I_End loop
         if Of_Username = Dereference(I).Username then
            -- user is in the database
            if And_Password = Dereference(I).Password or Dereference(I).Password = To_Unbounded_String("") then
               -- username and password is ok; or account has not password
               Result := Dereference(I);
               return;
            else
               -- username is ok but the password is invalid
               raise Access_Not_Allowed;
            end if;
         end if;

         Increment(I);
      end loop;

      raise Access_Not_Allowed;
      -- user is not in the database or invalid password
   end;

   -------------------
   -- Get_File_Tree --
   -------------------

   function Is_In_List (
         A_List                    : Shared_Dir_Lists.List; 
         Directory                 : Unbounded_String;      
         Filesystem_Case_Sensitive : Boolean                ) 
     return Boolean is 
      I     :          Shared_Dir_Lists.Iterator := Get_Begin (A_List);  
      I_End : constant Shared_Dir_Lists.Iterator := Get_End (A_List);  
   begin
      while I /= I_End loop

         if Filesystem_Case_Sensitive then
            --
            -- Un*x version of filesystem
            if Directory = Dereference(I).Name then
               return True;
            end if;
         else
            --
            -- NT version of filesystem
            if Insensitive_Compare(Directory, Dereference(I).Name) then
               return True;
            end if;
         end if;


         Increment(I);
      end loop;

      return False;
   end Is_In_List;

   -----------------------------------------------------------------------------------------

   function Get_File_Tree (
         Conf     : in     Ftp_Configuration_Type; 
         Username,                                 
         Password :        Unbounded_String        ) 
     return File_Tree_Type is 

      Result          : File_Tree_Type;  
      Account         : User_Configuration_Type;  
      Matched_Mapping : Mapping_Lists.List;  
      I,  
      I_End           : Mapping_Lists.Iterator;  
   begin
      Get_Account(Conf, Account, Username, Password);

      if not Is_In_List(Account.Directories, Account.Root_Directory, Conf.Filesystem_Case_Sensitive) then
         raise Access_Not_Allowed;
         -- Root directory is not in the shared directory list and has no access _
         -- rights assigned.
      end if;

      --
      -- Create list of mapped directories, but include only those that has _
      --  access rights assigned.
      I := Get_Begin(Conf.Mapping);
      I_End := Get_End(Conf.Mapping);
      while I /= I_End loop
         if Is_In_List(Account.Directories, Dereference(I).Directory, Conf.Filesystem_Case_Sensitive) then
            Push_Back(Dereference(I), Matched_Mapping);
         end if;
         Increment(I);
      end loop;

      Construct(Result, Account.Directories, Account.Root_Directory, Matched_Mapping, Conf.Filesystem_Case_Sensitive);

      return Result;

   end Get_File_Tree;

   --------------
   -- Get_Port --
   --------------

   function Get_Port (
         Conf : Ftp_Configuration_Type ) 
     return Port_Type is 
   begin
      Test_Loaded(Conf);
      return Conf.Port;
   end Get_Port;

   --
   -- Get_ip
   --

   function Get_Ip (
         Conf : Ftp_Configuration_Type ) 
     return Inet_Addr_Type is 
   begin
      Test_Loaded(Conf);
      return Conf.Ip;
   end;

   -------------------------
   -- Load_From_Directory --
   -------------------------
   -- and its slaves Load_Accounts, Load_mapping, Load_Root and ...

   procedure Load_Accounts (
         Accounts : in out User_Conf_Lists.List; 
         Filename :        Unbounded_String      ) is 

      Invalid_File_Format : exception;  
      File                : File_Type;  
      Line,  
      Key                 : Unbounded_String;  
   begin
      Empty(Accounts); -- defensive
      Open(File, In_File, To_String(Filename));

      Line := Get_Line(File);
      -- this is outside the user_loop because the "username" line is read when
      -- we strikes on another user section

      User_Loop :  loop
         -- reading the user accounts
         declare
            Current_User_Configuration : User_Configuration_Type;  
            --this will be added to the list in each iteration  
         begin
            Key := Get_Key(Line);

            if not Insensitive_Compare(Key, "username") then
               --
               --raise Invalid_File_Format;
               Raise_Exception(Invalid_File_Format'Identity, """username"" key expected, found """ & To_String(Key));
            end if;

            Current_User_Configuration.Username := Get_Value(Line);

            Line := Get_Line(File);
            Key := Get_Key(Line);

            if not Insensitive_Compare(Key, "password") then
               --
               --raise Invalid_File_Format;
               Raise_Exception(Invalid_File_Format'Identity, """password"" key expected, found """ & To_String(Key));
            end if;

            Current_User_Configuration.Password := Get_Value(Line);

            --
            -- reading root directory
            Line := Get_Line(File);
            Key := Get_Key(Line);

            if not Insensitive_Compare(Key, "root_dir") then
               --
               --raise Invalid_File_Format;
               Raise_Exception(Invalid_File_Format'Identity, """root_dir"" key expected, found """ & To_String(Key));
            end if;

            Current_User_Configuration.Root_Directory:= Normalize_Directory_Name(Get_Value(Line));

            Directory_Loop : loop
               --
               -- reading directories (and its access rights) for current user
               -- the root directory must be included in this list as well to recognise its rights
               declare
                  Current_Shared_Directory : Shared_Directory_Type;  
               begin
                  Line := Get_Line(File);

                  if Line = To_Unbounded_String("") then
                     -- we are at end of file => DONE
                     Push_Back(Current_User_Configuration, Accounts);
                     exit User_Loop;
                  end if;

                  Key := Get_Key(Line);

                  if Insensitive_Compare(Key, "username") then
                     -- we striked to another user entry, so save current and exit;
                     Push_Back(Current_User_Configuration, Accounts);
                     exit Directory_Loop;
                  end if;

                  if not Insensitive_Compare(Key, "physical_dir") then
                     --
                     --raise Invalid_File_Format;
                     Raise_Exception(Invalid_File_Format'Identity, """physical_dir"" key expected, found """ & To_String(Key));
                  end if;

                  Current_Shared_Directory.Name := Normalize_Directory_Name(Get_Value(Line));

                  --
                  -- reading access rights for loaded directory name
                  Line := Get_Line(File);
                  Key := Get_Key(Line);

                  if not Insensitive_Compare(Key, "rights") then
                     --
                     --raise Invalid_File_Format;
                     Raise_Exception(Invalid_File_Format'Identity, """rights"" key expected, found """ & To_String(Key));
                  end if;

                  Set(Current_Shared_Directory.Rights, Get_Value(Line));

                  --
                  -- Save the directory (with its access rights) to the current user's account
                  Push_Back(Current_Shared_Directory, Current_User_Configuration.Directories);
               end;
            end loop Directory_Loop;
         end;
      end loop User_Loop;

      Close(File);
   exception
      when Event : Invalid_File_Format =>
         Close(File);
         --
         -- raise Configuration_Not_Loaded;
         Raise_Exception(Configuration_Not_Loaded'Identity, Exception_Message(Event));
   end Load_Accounts;

   ---------------------------------------------------------------------------------

   procedure Load_Mapping (
         Mapping  : in out Mapping_Lists.List; 
         Filename :        Unbounded_String    ) is 

      File                : File_Type;  
      Line,  
      Key                 : Unbounded_String;  
      Invalid_File_Format : exception;  
   begin
      Empty(Mapping); -- defensive
      Open(File, In_File, To_String(Filename));

      loop
         declare
            Current_Mapping : Mapping_Type;  
         begin
            Line := Get_Line(File);

            if Line = To_Unbounded_String("") then
               -- end of file
               exit;
            end if;

            Key := Get_Key(Line);

            if not Insensitive_Compare(Key, "physical_dir") then
               --
               --raise Invalid_File_Format;
               Raise_Exception(Invalid_File_Format'Identity, """physical_dir"" key expected, found """ & To_String(Key) & """");
            end if;

            Current_Mapping.Directory := Normalize_Directory_Name(Get_Value(Line));

            Line := Get_Line(File);
            Key := Get_Key(Line);

            if not Insensitive_Compare(Key, "mapped_as") then
               --
               --raise Invalid_File_Format;
               Raise_Exception(Invalid_File_Format'Identity, """mapped_as"" key expected, found """ & To_String(Key) & """");
            end if;

            Current_Mapping.Alias := Get_Value(Line);

            Push_Back(Current_Mapping, Mapping);
         end;
      end loop;

      Close(File);
   exception
      when Event : Invalid_File_Format =>
         Close(File);
         --
         -- raise Configuration_Not_Loaded;
         Raise_Exception(Configuration_Not_Loaded'Identity, Exception_Message(Event));
   end Load_Mapping;


   ---------------------------------------------------------------------------------

   procedure Load_Root (
         Conf     : in out Ftp_Configuration_Type; 
         Filename :        Unbounded_String        ) is 

      Invalid_File_Format : exception;  
      Line,  
      Key                 : Unbounded_String;  
      File                : File_Type;  
   begin
      --
      -- defensive settings
      Conf.Ip := No_Inet_Addr;
      Conf.Port := No_Port;
      Conf.Filesystem_Case_Sensitive := True;
      Open(File, In_File, To_String(Filename));

      --
      -- reading ip
      Line := Get_Line(File);
      Key := Get_Key(Line);

      if not Insensitive_Compare(Key, "ip") then
         --
         --raise Invalid_File_Format;
         Raise_Exception(Invalid_File_Format'Identity, """ip"" key expected, found """ & To_String(Key) & """");
      end if;

      Conf.Ip := Inet_Addr(+Get_Value(Line));

      --
      -- reading port
      Line := Get_Line(File);
      Key := Get_Key(Line);

      if not Insensitive_Compare(Key, "port") then
         --
         --raise Invalid_File_Format;
         Raise_Exception(Invalid_File_Format'Identity, """port"" key expected, found """ & To_String(Key) & """");
      end if;

      Conf.Port := Port_Type(Integer'Value(To_String(Get_Value(Line))));
      -- raises C_E if not valid integer

      --
      -- Reading filesystem option
      Line := Get_Line(File);
      Key := Get_Key(Line);

      if not Insensitive_Compare(Key, "fs_case_sensitive") then
         --
         --raise Invalid_File_Format;
         Raise_Exception(Invalid_File_Format'Identity, """fs_case_sensitive"" key expected, found """ & To_String(Key));
      end if;

      if Get_Value(Line) = "1" then
         Conf.Filesystem_Case_Sensitive := True;
      elsif Get_Value(Line) = "0" then
         Conf.Filesystem_Case_Sensitive := False;
      else
         raise Invalid_File_Format;
      end if;

      Close(File);
   exception
      when Event : Invalid_File_Format | Constraint_Error =>
         Close(File);
         --
         -- raise Configuration_Not_Loaded;
         Raise_Exception(Configuration_Not_Loaded'Identity, Exception_Message(Event));
   end Load_Root;

   ---------------------------------------------------------------------------------

   --
   -- The guru itself
   --

   procedure Load_From_Directory (
         Conf      : in out Ftp_Configuration_Type; 
         Directory :        Unbounded_String        ) is 
   begin
      Load_Accounts(Conf.Accounts, Compose_Path(Directory, To_Unbounded_String("accounts.txt")));
      Load_Mapping(Conf.Mapping, Compose_Path(Directory, To_Unbounded_String("mapping.txt")));
      Load_Root(Conf, Compose_Path(Directory, To_Unbounded_String("root.txt")));

      Conf.Loaded := True;

   exception
      when Name_Error => -- the given directory is not valid (does not contains appropriate files)
         raise Configuration_Not_Loaded;
   end Load_From_Directory;

   -----------------
   -- Verify_User --
   -----------------

   procedure Verify_User (
         Conf     : Ftp_Configuration_Type; 
         Username : Unbounded_String        ) is 

      I     :          User_Conf_Lists.Iterator := Get_Begin (Conf.Accounts);  
      I_End : constant User_Conf_Lists.Iterator := Get_End (Conf.Accounts);  
   begin
      Test_Loaded(Conf);

      while I /= I_End loop
         if Username = Dereference(I).Username then
            -- user is in the database
            return;
         end if;

         Increment(I);
      end loop;

      raise Access_Not_Allowed;
      -- user is not in the database
   end Verify_User;

   -----------------
   -- Verify_User --
   -----------------

   procedure Verify_User (
         Conf     : Ftp_Configuration_Type; 
         Username,                          
         Password : Unbounded_String        ) is 
      To_Throw : User_Configuration_Type;  
   begin
      Get_Account(Conf, To_Throw, Username, Password); --raises Access_Not_Allowed when fail
   end Verify_User;

   --
   -- Log
   --
   --      type Ftp_Configuration_Type is new Controlled with 
   --      record 
   --         Loaded                    : Boolean;  
   --         Ip                        : Inet_Addr_Type;  
   --         Port                      : Port_Type;  
   --         Filesystem_Case_Sensitive : Boolean;  
   --         Accounts                  : User_Conf_Lists.List;  
   --         Mapping                   : Mapping_Lists.List;  
   --      end record;


   procedure Log (
         Conf : Ftp_Configuration_Type ) is 
   begin
      Log("Loaded configuration:");
      Log("IP = " & Image(Conf.Ip));
      Log("Port = " & Integer'Image(Integer(Conf.Port)));
      Log("Filesystem case sensitive = " & Boolean'Image(Conf.Filesystem_Case_Sensitive));
      Log("List of user accounts = ");

      declare
         use User_Conf_Lists;
         I     :          User_Conf_Lists.Iterator := Get_Begin (Conf.Accounts);  
         I_End : constant User_Conf_Lists.Iterator := Get_End (Conf.Accounts);  
      begin
         while I /= I_End loop
            Log(Dereference(I));
            Increment(I);
         end loop;
      end;

      Log("List of mapped directories = ");

      declare
         use Mapping_Lists;
         I     :          Mapping_Lists.Iterator := Get_Begin (Conf.Mapping);  
         I_End : constant Mapping_Lists.Iterator := Get_End (Conf.Mapping);  
      begin
         while I /= I_End loop
            Log(Dereference(I));
            Increment(I);
         end loop;
      end;
   end;

   --
   -- private stuff
   --

   -----------------
   -- Test_Inited --
   -----------------

   procedure Test_Loaded (
         Conf : Ftp_Configuration_Type ) is 
   begin
      if not Conf.Loaded then
         raise Configuration_Not_Loaded;
      end if;
   end Test_Loaded;

   --
   -- Controlled stuff
   --

   procedure Initialize (
         Conf : in out Ftp_Configuration_Type ) is 
   begin
      Conf.Loaded:= False;
      Conf.Filesystem_Case_Sensitive := True;
      Conf.Port := No_Port;
   end;


end Ftp_Configurations;

