with Ada.Unchecked_Deallocation,
   Ada.Exceptions,
   Ftp_Server.Session,
   Paskma.Logger,
   Task_Id_Types,
   Paskma.Lists,
   --Ada.Text_Io,
   Gnat.Sockets;

use Ftp_Server.Session,
   Ada.Exceptions,
   Paskma.Logger,
   Task_Id_Types,
   --Ada.Text_Io,
   Gnat.Sockets;

package body Ftp_Server is
   package Session_Task_Lists is new Paskma.Lists(Session_Task_Access);
   use Session_Task_Lists;

   ----------------------
   -- Server_Task_Type --
   ----------------------
   -- and its private procedures

   --
   -- Release_Terminated
   -- 
   -- Releases old session tasks descriptors

   procedure Release_Terminated (
         Session_Tasks : in out List ) is 

      Task_To_Release :          Session_Task_Access;  
      I               :          Iterator            := Get_Begin (Session_Tasks);  
      I_End           : constant Iterator            := Get_End (Session_Tasks);  

      procedure Release is 
      new Ada.Unchecked_Deallocation(Session_Task_Type,
         Session_Task_Access);
   begin

      while I /= I_End loop -- passes whole list
         if Dereference(I)'Terminated then
            -- only if the task is successfuly done we can release the descriptor
            Task_To_Release := Dereference(I);
            -- we cannot release directly the item of the list
            Release(Task_To_Release);

            Remove(I, Session_Tasks);
            -- after "Remove", the iterator is incremented and is valid again
         else
            Increment(I);
         end if;
      end loop;
   end;

   --
   -- Create_New_Task
   --
   -- internal procedure for Server_Task_Type
   -- called by Server_Task_Type to create (and configure) a new session task

   procedure Create_New_Task (
         New_Task :    out Session_Task_Access; 
         Parent   :        Server_Task_Access;  
         -- access to parent will be useful for creating new session tasks
         A_Socket : Socket_Type; 
         -- new session task will wait for socket accept on this port
         A_Configuration : Ftp_Configuration_Type;   
         A_Statistics    : Statistics_Manager_Access ) is 
      -- protected object to hold information for user 
   begin
      New_Task :=  new Session_Task_Type;

      New_Task.Start(
         A_Configuration         => A_Configuration,
         Accept_On_Socket        => A_Socket,
         Parent_For_New_Sessions => Parent_Shortcut_Type'(Server_Task =>
            Parent),
         Some_Task_Id            => Unique_Task_Id,
         Store_Statiscs_Information  => A_Statistics);

   end;

   --
   -- The Server_Task body itself
   --

   task body Server_Task_Type is
      Listening_Socket              :          Socket_Type;  
      Listening_Address             :          Sock_Addr_Type;  
      Self                          :          Server_Task_Access;  
      Configuration                 :          Ftp_Configuration_Type;  
      Accepting_Task                :          Session_Task_Access;  
      Session_Tasks                 :          List;  
      Statistics                    :          Statistics_Manager_Access;  
      Name_Completer_For_Statistics :          Name_Completer_Task_Type;  
      Max_Session_Task_Count        : constant                           := 255;  
   begin
      accept Start (
            Load_Config_From             : Ftp_Configuration_Type;    
            Store_Statistics_Information : Statistics_Manager_Access; 
            We_Need_Self                 : Server_Task_Access         ) do 

         Configuration := Load_Config_From;
         Statistics := Store_Statistics_Information;
         Self := We_Need_Self;
      end Start;

      --
      -- Some stuff to initialize sockets
      Listening_Address.Addr := Get_Ip(Configuration);
      Listening_Address.Port := Get_Port(Configuration);
      Create_Socket(Listening_Socket);
      Bind_Socket(Listening_Socket, Listening_Address);
      Listen_Socket(Listening_Socket);
      Log("Listening on address: " & Image(Listening_Address));
      Name_Completer_For_Statistics.Start(Statistics, 10.0);

      Log("Creating basic session..");
      Create_New_Task(Accepting_Task, Self, Listening_Socket,
         Configuration, Statistics);
      Push_Back(Accepting_Task, Session_Tasks);
      -- create basic accepting task, next will be created by Create_New_Session accept

      --
      -- The main infinite accept loop for creating new sessions
      loop
         select
            accept Create_New_Session do 
               -- this accept is not called directly by session task, but via private procedure "Create_New_Session"
               Log("Creating session...");
               Create_New_Task(Accepting_Task, Self, Listening_Socket,
                  Configuration, Statistics);

               Push_Back(
                  Item        => Accepting_Task, 
                  To_The_List => Session_Tasks);
               if Length(Session_Tasks) > Max_Session_Task_Count then
                  Release_Terminated(Session_Tasks);
                  -- there were too meny tasks created, try release some of them which are completed
               end if;
            end Create_New_Session;
         or
            accept Stop do 
               -- usualy called by public procedure "Stop"
               Log("The server task is closing the socket...");
               Close_Socket(Listening_Socket);
               -- this will cause the end of currently accepting session task
               Name_Completer_For_Statistics.Stop;
            end;
            exit;
         end select;
      end loop;

   exception
      when Constraint_Error =>
         Log("unexpected C_E");
      when Socket_Error =>
         Log(
            "Unexpected socket_error, probably another server is running");
         abort Name_Completer_For_Statistics;
      -- if we are here, the name_completer is probably waiting for Start entry
      when Event : others =>
         Log("Unexpected exception in Server_Task " & Exception_Information(Event));

   end Server_Task_Type;

   -----------
   -- Start --
   -----------

   procedure Start (
         Server_Context :    out Server_Context_Type; 
         Config_Path    :        Unbounded_String;    
         Success        :    out Boolean;             
         Message        :    out Unbounded_String     ) is 

      Configuration : Ftp_Configuration_Type;  
   begin
      Load_From_Directory(Configuration, Config_Path);
      Log(Configuration);
      Server_Context.Server_Task := new Server_Task_Type;
      Server_Context.Server_Task.Start(
         Configuration,
         Server_Context.Information'Unchecked_Access,
         -- this is realy not safe. Especialy when exiting the server
         --  (the ftp-pi tasks are not stopped and they can use blind pointer in excterme case)
         Server_Context.Server_Task);

      Log("Start procedure finished.");
      Success := True;
      Message := To_Unbounded_String("Server has started.");
   exception
      when Event : others =>
         Success := False;
         Message := To_Unbounded_String(Exception_Information(Event));

   end Start;

   ----------
   -- Stop --
   ----------
   -- Ater this procedure is executed, the ftp-pi tasks can be still alive
   -- Especially the Inforation field of server context should be accessible
   -- until these posthumos-tasks are alive.

   procedure Stop (
         Server_Context : in out Server_Context_Type; 
         Message        :    out Unbounded_String     ) is 
      procedure Release is 
      new Ada.Unchecked_Deallocation(Server_Task_Type, Server_Task_Access);
   begin
      Message := To_Unbounded_String("Everything OK.");
      begin

         if Server_Context.Server_Task /= null then
            -- server is running (or was running)
            Server_Context.Server_Task.Stop;
            -- this might cause tasing_error exception
            while not Server_Context.Server_Task'Terminated loop
               delay 0.3;
            end loop;
         else
            Message := To_Unbounded_String("Server not started.");
         end if;

      exception
         when Tasking_Error =>
            -- server was not running
            Message := To_Unbounded_String(
               "Cannot stop server, it is probably already down.");
            Log(Message);
      end;

      Release(Server_Context.Server_Task);

   exception
      when Event : others =>
         Message := To_Unbounded_String(Exception_Information(Event));
         Log(Message);
   end Stop;

   --
   -- Get_Information
   --

   function Get_Information (
         Server_Context : Server_Context_Type ) 
     return Unbounded_String is 
   begin
      return Server_Context.Information.Get_Info;
   end;

   --
   -- Is_Alive
   --

   function Is_Alive (
         Server_Context : Server_Context_Type ) 
     return Boolean is 
      Result : Boolean;  
   begin
      if Server_Context.Server_Task = null then
         -- server was not initialized
         --Put_Line(
         --   "--Is_Alive: server was not initialized, returning false");
         return False;
      else
         Result := Server_Context.Server_Task'Terminated;
         if Result then
            --Put_Line("--Is_Alive: server is terminated, returning false");
            return False;
         else
            --Put_Line(
            --   "--Is_Alive: server is not terminated, returning false");
            return True;
         end if;

      end if;
   end;

   --
   -- Create_new_session
   --
   -- Private procedure which is called directly by session task (it is in child package)

   procedure Create_New_Session (
         Parent : Parent_Shortcut_Type ) is 
   begin
      Parent.Server_Task.Create_New_Session;
   end;

   --
   -- Init
   --

   procedure Init is 
   begin
      Gnat.Sockets.Initialize(True);
   end Init;

   --
   -- Finalize
   --

   procedure Finalize is 
   begin
      Gnat.Sockets.Finalize;
   end;
end Ftp_Server;

