--
-- @filename pftp.adb
-- @author Marek Paska
-- @date 20. 12. 02
-- @brief The main program file
--
-- History :
-- 20. 12. 02 AdaDoc documentation
--

with Ftp_Server,
   Ada.Text_Io,
   Ada.Strings.Unbounded,
   Ada.Calendar,
   Log_Flushers,
   Paskma.Time_Utilities,
   Paskma.Logger;

use Ftp_Server,
   Ada.Text_Io,
   Ada.Strings.Unbounded,
   Ada.Calendar,
   Paskma.Logger,
   Paskma.Time_Utilities,
   Log_Flushers;

--
-- Purpose: Starts the program. The main procedure
--
procedure Pftp is 
   The_Server   : Server_Context_Type;  
   Command      : Unbounded_String;  
   Temp_Command : String (1 .. 256);  
   Command_Last : Natural;  

   Success : Boolean;  
   Message : Unbounded_String;  

   Log_Flusher : Log_Flusher_Type;  
   Start_Time  : Time;  
begin
   begin
      Init_Log(To_String("logs/_protocol_" & Get_Archive_Time_Stamp & ".log"));
   exception
      when others =>
         Put_Line("Error opening log file. Directory ""./logs"" probably not accessible.");
   end;
   Log("Paskma FTP Server started.");
   Log_Flusher.Start(10.0); -- flush the log buffer every 10 seconds
   Ftp_Server.Init; -- initializes things like sockets

   --
   -- Starts the default server. The configuration is stored in the directory named "default" in
   -- the server's home directory. The pftp is designed to run various server processes on  different
   -- ports with different configurations in varions directories. But THIS silly administration
   -- console is designed to administrate ONE ftp-server-task with configuration in "default" directory.
   --

   Put_Line("Starting the server process. Loading configuration from directory ""default"".");
   Start(The_Server, To_Unbounded_String("default"), Success, Message);

   if Success then
      Put_Line("Paskma FTP Server started...");
      delay 0.5;
      if not Is_Alive(The_Server) then
         Put_Line("Cannot create listen socket. Probably another server is running.");
         Put_Line("Ensure server status again and/or try restart.");
      end if;
   else
      Put_Line("Server error: ");
      Put_Line(To_String(Message));
   end if;

   Start_Time := Clock;

   loop
      Put(">");
      Get_Line(Temp_Command, Command_Last);
      Command := To_Unbounded_String(Temp_Command(1..Command_Last));

      if Command = "exit" then
         --
         -- Exit
         --
         Put_Line("Stoping the server...");
         Stop(The_Server, Message);
         Put_Line("Server stopped with message: " & To_String(Message));
         Log("Program ended by user.");
         Log_Flusher.Stop;
         Close_Log;
         exit;
      elsif Command = "traffic" then
         --
         -- Trafic
         --
         Put_Line(To_String(Get_Information(The_Server)));
      elsif Command = "help" then
         --
         -- Help
         --
         New_Line;
         Put_Line(" Implemented commands:");
         Put_Line("   traffic");
         Put_Line("   uptime");
         Put_Line("   status");
         Put_Line("   restart");
         Put_Line("   help");
         Put_Line("   exit");
         New_Line;
      elsif Command = "status" then
         --
         -- Status
         --
         if Is_Alive(The_Server) then
            Put_Line("Server task is running.");
         else
            Put_Line("Server task is NOT running.");
         end if;
      elsif Command = "restart" then
         --
         -- Restart
         --

         --
         -- Stop
         Put_Line("Stoping the server...");
         Stop(The_Server, Message);
         Put_Line("Server stopped with message: " & To_String(Message));

         --
         -- Start
         Put_Line("Starting the server process. Configuration loading from directory ""default""");
         Start(The_Server, To_Unbounded_String("default"), Success, Message);

         if Success then
            Put_Line("Paskma FTP Server started...");
            delay 0.5;
            if not Is_Alive(The_Server) then
               Put_Line("Cannot create listen socket. Probably another server is running.");
               Put_Line("Ensure server status again and/or try restart.");
            end if;
         else
            Put_Line("Server error: ");
            Put_Line(To_String(Message));
         end if;
      elsif Command = "uptime" then
         --
         -- Uptime
         --
         declare
            Uptime,  
            Days_Uptime,  
            Hours_Uptime,  
            Minutes_Uptime,  
            Seconds_Uptime : Natural;  
         begin
            Uptime := Natural(Clock -  Start_Time);
            Days_Uptime := Uptime / (3600*24);
            Hours_Uptime := Uptime / 3600 mod 24;
            Minutes_Uptime := Uptime / 60  mod 60;
            Seconds_Uptime := Uptime mod 60;

            Put_Line("Uptime is " & Natural'Image(Days_Uptime) & " days " & Natural'Image(Hours_Uptime) & " hours "
               & Natural'Image(Minutes_Uptime) & " minutes " & Natural'Image(Seconds_Uptime) & " seconds.");
         end;
      elsif Command = Null_Unbounded_String then
         --
         -- just <enter>
         --
         null;
      else
         --
         -- unknown command
         --
         Put_Line("Unknown command - " & To_String(Command));
         Put_Line("Type ""help"" for information.");
         New_Line;
      end if;
   end loop;
   Ftp_Server.Finalize;
end;
